==> store-2019.9.9-py3-none-any.whl: store/database.py <==
import json
import os
import uuid
from copy import copy
from datetime import datetime

from pony.orm import (Database, Json, PrimaryKey, Required, commit, count,
                      db_session, delete, desc, select)
from store.parser import parse
try:
    from cerberus import Validator
    SCHEMA_CHECK = True
except Exception:
    SCHEMA_CHECK = False


class StoreException(Exception):
    pass

class StoreMetas:
    def __init__(self, elems, store=None):
        if not elems:
            elems = []
        self.elems = [StoreMeta(elem, store=store) for elem in elems]

    def __str__(self):
        return '\n'.join([str(elem) for elem in self.elems])

    def __len__(self):
        return len(self.elems)

    @db_session
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.elems[key]
        if isinstance(key, slice):
            return self.elems[key]
        return [elem[key] for elem in self.elems]

    @db_session
    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.elems[key] = value
            return
        for elem in self.elems:
            elem[key] = value


    @db_session
    def __getattribute__(self, key):
        if key in ['elems'] or key.startswith('_'):
            return object.__getattribute__(self, key)

        if key in ['store', 'id', 'key', 'value', 'create', 'update']:
            return [elem.__getattribute__(key) for elem in self.elems]

        return [elem[key] for elem in self.elems]


class StoreMeta:
    def __init__(self, elem, store=None):
        self.store = store
        self.id = elem.id
        self.key = elem.key
        self.value = elem.value
        self.create = elem.create.strftime("%Y-%m-%dT%H:%M:%S")
        self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")

    def __str__(self):
        return "id: {}, key: {}, value: {}, create: {}, update: {}".format(self.id, self.key, self.value, self.create, self.update)

    @db_session
    def __assign__(self, value):
        self.store.validate(value)
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise StoreException('elem not found')
        else:
            elem.value = value
            elem.update = datetime.utcnow()

            self.value = elem.value
            self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")

    @db_session
    def __setattr__(self, key, value):
        if key in ['store', 'id', 'key', 'value', 'create', 'update'] or key.startswith('_'):
            return super().__setattr__(key, value)
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise StoreException('elem not found')
        else:
            if isinstance(elem.value, dict):
                elem.value[key] = value
                elem.update = datetime.utcnow()

                self.value = elem.value
                self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")
            else:
                raise StoreException('value not dict!')

    @db_session
    def __getattribute__(self, key):
        if key in ['store', 'id', 'key', 'value', 'create', 'update'] or key.startswith('_'):
            return object.__getattribute__(self, key)

        elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            if isinstance(elem.value, dict):
                print(self.store, '......')
                return elem.value.get(key)

    @db_session
    def __setitem__(self, key, value):
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise StoreException('elem not found')
        else:
            if isinstance(elem.value, dict) or \
               (isinstance(key, int) and isinstance(elem.value, list)):
                elem.value[key] = value
                elem.update = datetime.utcnow()

                self.value = elem.value
                self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")
            else:
                raise StoreException('value not dict!')

    @db_session
    def __getitem__(self, key):
        elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            if isinstance(elem.value, dict) or \
               (isinstance(key, int) and isinstance(elem.value, list)) or \
               (isinstance(key, int) and isinstance(elem.value, str)):
                if isinstance(key, int):
                    return elem.value[key]
                else:
                    return elem.value.get(key)
class Store(object):
    _safe_attrs = ['store', 'database', 'tablename', 
                   'begin', 'end', 'order', 
                   'add', 'register_attr', 'slice', 'adjust_slice', 'provider',
                   'query_key', 'count', 'desc', 'asc',
                   'provider', 'user', 'password', 'host', 'port', 'database', 'filename',
                   'schema', 'validate'
                   ]

    provider = 'sqlite'
    user = 'test'
    password = 'test'
    host = 'localhost'
    port = 5432
    database = 'test'
    filename = 'database.sqlite'
    order = 'desc'
    schema = None
    begin = None
    end = None
    

    def __init__(self,
                 provider=None, user=None, password=None,
                 host=None, port=None, database=None, filename=None,
                 begin=None, end=None, order=None,
                 schema = None, validate=None
                 ):
        self.provider = provider or self.provider
        self.filename = filename or self.filename
        self.user = user or self.user
        self.password = password or self.password
        self.host = host or self.host
        self.port = port or self.port
        self.database = database or self.database
        self.schema = schema or self.schema
        self.begin = begin or self.begin
        self.end = end or self.end
        self.order = order or self.order

        if provider == 'sqlite':
            if not self.filename.startswith('/'):
                self.filename = os.getcwd()+'/' + self.filename

            self.database = Database(
                provider=self.provider, 
                filename=self.filename, 
                create_db=True)
        else:
            self.database = Database(
                provider=self.provider, 
                user=self.user, 
                password=self.password,
                host=self.host, 
                port=self.port, 
                database=self.database)

        # self.provider = provider

        self.tablename = self.__class__.__name__

        schema = dict(
            id=PrimaryKey(int, auto=True),
            create=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
            update=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
            key=Required(str, index=True, unique=True),
            value=Required(Json, volatile=True)
        )

        self.store = type(self.tablename, (self.database.Entity,), schema)
        self.database.generate_mapping(create_tables=True, check_tables=True)


    def validate(self, value, extra=None):
        if SCHEMA_CHECK and self.schema:
            validator = Validator()
            validator.schema = self.schema
            r = validator.validate(value)
            if not r:
                if extra:
                    raise StoreException(f'{validator.errors}, extra: {extra}')
                raise StoreException(validator.errors)

    def slice(self, begin, end):
        self.begin, self.end = begin, end

    def desc(self):
        self.order = 'desc'

    def asc(self):
        self.order = 'asc'

    @staticmethod
    def register_attr(name):
        if isinstance(name, str) and name not in Store._safe_attrs:
            Store._safe_attrs.append(name)

    @db_session
    def __setattr__(self, key, value):
        if key in Store._safe_attrs or key.startswith('_'):
            return super().__setattr__(key, value)

        self.validate(value)
        item = select(e for e in self.store if e.key == key).first()
        if item is None:
            self.store(key=key, value=value)
        else:
            item.value = value
            item.update = datetime.utcnow()

    @db_session
    def __getattribute__(self, key):
        if key in Store._safe_attrs or key.startswith('_'):
            return object.__getattribute__(self, key)

        elem = select(e for e in self.store if e.key == key).first()
        if elem:
            self.validate(elem.value)
            return StoreMeta(elem, store=self.store)
        return None

    @db_session
    def count(self, key):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        return elems.count()

    @db_session
    def __getitem__(self, key):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=False)
        for elem in elems:
            self.validate(elem.value, extra=elem.key)
        return StoreMetas(elems, store=self.store)


    @db_session
    def __setitem__(self, key, value):

        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)
        
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order_by == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=True)
        if elems:
            now = datetime.utcnow()
            for elem in elems:
                elem.value = value
                elem.update = now
        else:
            if key.isidentifier():
                return self.__setattr__(key, value)
            raise Exception('Not Implemented!')
        return

        

            

    @db_session
    def __delitem__(self, key):
        if isinstance(key, slice):
            raise Exception('not implemented!')
        elif isinstance(key, tuple):
            key = '.'.join(key)
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order_by == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        if elems:
            for elem in elems:
                self.validate(elem.value)
                elem.delete()
        return
       

    @db_session
    def __delattr__(self, key):
        delete(e for e in self.store if e.key == key)


    @db_session
    def add(self, value, key=None):
        self.validate(value)
        hex = uuid.uuid1().hex
        key = "_{}".format(hex) if not isinstance(key, str) else key
        self.store(key=key, value=value)
        return key

    @db_session
    def query_key(self, key, for_update=False):
        elem = None
        if for_update:
            elem = select(e for e in self.store if e.id == self.id).for_update().first()
        else:
            elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            self.validate(elem.value)
            return StoreMeta(elem, store=self.store)

    def adjust_slice(self, elems, for_update=False):
        if for_update:
            elems = elems.for_update()
        begin, end = self.begin, self.end
        if begin and end:
            # pony doesn't support step here
            if begin < 0:
                begin = len(self) + begin
            if end < 0:
                end = len(self) + end
            if begin > end:
                begin, end = end, begin
            elems = elems[begin:end]
        elif begin:
            if begin < 0:
                begin = len(self) + begin
            elems = elems[begin:]
        elif end:
            if end < 0:
                end = len(self) + end
            elems = elems[:end]
        else:
            elems = elems[:]
        return elems

    @db_session
    def __len__(self):
        return count(e for e in self.store)

==> store-2019.9.10-py3-none-any.whl: store/database.py <==
import json
import os
import uuid
from copy import copy
from datetime import datetime

from pony.orm import (Database, Json, PrimaryKey, Required, commit, count,
                      db_session, delete, desc, select)
from store.parser import parse
try:
    from cerberus import Validator
    SCHEMA_CHECK = True
except Exception:
    SCHEMA_CHECK = False


class StoreException(Exception):
    pass

class StoreMetas:
    def __init__(self, elems, store=None):
        if not elems:
            elems = []
        self.elems = [StoreMeta(elem, store=store) for elem in elems]

    def __str__(self):
        return '\n'.join([str(elem) for elem in self.elems])

    def __len__(self):
        return len(self.elems)

    @db_session
    def __getitem__(self, key):
        if isinstance(key, int):
            return self.elems[key]
        if isinstance(key, slice):
            return self.elems[key]
        return [elem[key] for elem in self.elems]

    @db_session
    def __setitem__(self, key, value):
        if isinstance(key, int):
            self.elems[key] = value
            return
        for elem in self.elems:
            elem[key] = value


    @db_session
    def __getattribute__(self, key):
        if key in ['elems'] or key.startswith('_'):
            return object.__getattribute__(self, key)

        if key in ['store', 'id', 'key', 'value', 'create', 'update', 'meta']:
            return [elem.__getattribute__(key) for elem in self.elems]

        return [elem[key] for elem in self.elems]


class StoreMeta:
    def __init__(self, elem, store=None):
        self.store = store
        self.id = elem.id
        self.key = elem.key
        self.value = elem.value
        self.create = elem.create.strftime("%Y-%m-%dT%H:%M:%S")
        self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")

    def __str__(self):
        return "id: {}, key: {}, value: {}, create: {}, update: {}".format(self.id, self.key, self.value, self.create, self.update)

    @db_session
    def __assign__(self, value):
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise Exception('elem not found')
        else:
            elem.value = value
            elem.update = datetime.utcnow()

            self.value = elem.value
            self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")

    @db_session
    def __setattr__(self, key, value):
        if key in ['store', 'id', 'key', 'value', 'create', 'update'] or key.startswith('_'):
            return super().__setattr__(key, value)
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise Exception('elem not found')
        else:
            if key == 'meta':
                elem.meta = value
                elem.update = datetime.utcnow()
            elif isinstance(elem.value, dict):
                elem.value[key] = value
                elem.update = datetime.utcnow()

                self.value = elem.value
                self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")
            else:
                raise Exception('value not dict!')

    @db_session
    def __getattribute__(self, key):
        if key in ['store', 'id', 'key', 'value', 'create', 'update'] or key.startswith('_'):
            return object.__getattribute__(self, key)

        elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            if key=='meta':
                return elem.meta
            if isinstance(elem.value, dict):
                return elem.value.get(key)

    @db_session
    def __setitem__(self, key, value):
        elem = select(e for e in self.store if e.id == self.id).for_update().first()
        if elem is None:
            raise Exception('elem not found')
        else:
            if isinstance(elem.value, dict) or \
               (isinstance(key, int) and isinstance(elem.value, list)):
                elem.value[key] = value
                elem.update = datetime.utcnow()

                self.value = elem.value
                self.update = elem.update.strftime("%Y-%m-%dT%H:%M:%S")
            else:
                raise Exception('value not dict!')

    @db_session
    def __getitem__(self, key):
        elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            if isinstance(elem.value, dict) or \
               (isinstance(key, int) and isinstance(elem.value, list)) or \
               (isinstance(key, int) and isinstance(elem.value, str)):
                if isinstance(key, int):
                    return elem.value[key]
                else:
                    return elem.value.get(key)


class Store(object):
    _safe_attrs = ['store', 'database', 'tablename', 
                   'begin', 'end', 'order', 
                   'add', 'register_attr', 'slice', 'adjust_slice', 'provider',
                   'query_key', 'count', 'desc', 'asc',
                   'provider', 'user', 'password', 'host', 'port', 'database', 'filename',
                   'schema', 'validate', 'version', 'model'
                   ]

    provider = 'sqlite'
    user = 'test'
    password = 'test'
    host = 'localhost'
    port = 5432
    database = 'test'
    filename = 'database.sqlite'
    order = 'desc'
    schema = None
    begin = None
    end = None
    version = "meta"
    model = None
    

    def __init__(self,
                 provider=None, user=None, password=None,
                 host=None, port=None, database=None, filename=None,
                 begin=None, end=None, order=None,
                 schema = None, validate=None, version="meta", model=None
                 ):
        self.provider = provider or self.provider
        self.filename = filename or self.filename
        self.user = user or self.user
        self.password = password or self.password
        self.host = host or self.host
        self.port = port or self.port
        self.database = database or self.database
        self.schema = schema or self.schema
        self.begin = begin or self.begin
        self.end = end or self.end
        self.order = order or self.order
        self.version = version or self.version
        self.model = model or self.model

        if provider == 'sqlite':
            if not self.filename.startswith('/'):
                self.filename = os.getcwd()+'/' + self.filename

            self.database = Database(
                provider=self.provider, 
                filename=self.filename, 
                create_db=True)
        else:
            self.database = Database(
                provider=self.provider, 
                user=self.user, 
                password=self.password,
                host=self.host, 
                port=self.port, 
                database=self.database)

        self.tablename = self.__class__.__name__

        if not self.model:
            if not version:
                self.model = dict(
                    id=PrimaryKey(int, auto=True),
                    create=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
                    update=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
                    key=Required(str, index=True, unique=True),
                    value=Required(Json, volatile=True)
                )
            else:
                self.model = dict(
                    id=PrimaryKey(int, auto=True),
                    create=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
                    update=Required(datetime, sql_default='CURRENT_TIMESTAMP', default=lambda: datetime.utcnow()),
                    key=Required(str, index=True, unique=True),
                    value=Required(Json, volatile=True, default={}),
                    meta=Required(Json, volatile=True, default={})
                )

        self.store = type(self.tablename, (self.database.Entity,), self.model)
        self.database.generate_mapping(create_tables=True, check_tables=True)


    def validate(self, value, extra=None, meta=None):
        if SCHEMA_CHECK and self.schema:
            validator = Validator()
            if meta:
                schema_version = meta.get("schema_version")
                if schema_version:
                    validator.schema = self.schema.get(schema_version)
                else:
                    validator.schema = self.schema
            else:
                validator.schema = self.schema
            r = validator.validate(value)
            if not r:
                if extra:
                    raise StoreException(f'{validator.errors}, extra: {extra}')
                raise StoreException(validator.errors)

    def slice(self, begin, end):
        self.begin, self.end = begin, end

    def desc(self):
        self.order = 'desc'

    def asc(self):
        self.order = 'asc'

    @staticmethod
    def register_attr(name):
        if isinstance(name, str) and name not in Store._safe_attrs:
            Store._safe_attrs.append(name)

    @db_session
    def __setattr__(self, key, value):
        if key in Store._safe_attrs or key.startswith('_'):
            return super().__setattr__(key, value)

        self.validate(value)
        item = select(e for e in self.store if e.key == key).first()
        if item is None:
            self.store(key=key, value=value)
        else:
            item.value = value
            item.update = datetime.utcnow()

    @db_session
    def __getattribute__(self, key):
        if key in Store._safe_attrs or key.startswith('_'):
            return object.__getattribute__(self, key)

        elem = select(e for e in self.store if e.key == key).first()
        if elem:
            self.validate(elem.value)
            return StoreMeta(elem, store=self.store)
        return None

    @db_session
    def count(self, key):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        return elems.count()

    @db_session
    def __getitem__(self, key):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=False)
        for elem in elems:
            self.validate(elem.value, extra=elem.key, meta=elem.meta)
        return StoreMetas(elems, store=self.store)


    @db_session
    def __setitem__(self, key, value):

        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)
        
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order_by == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=True)
        if elems:
            now = datetime.utcnow()
            for elem in elems:
                elem.value = value
                elem.update = now
        else:
            if key.isidentifier():
                return self.__setattr__(key, value)
            raise Exception('Not Implemented!')
        return

        

            

    @db_session
    def __delitem__(self, key):
        if isinstance(key, slice):
            raise Exception('not implemented!')
        elif isinstance(key, tuple):
            key = '.'.join(key)
        filters = parse(key)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order_by == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        if elems:
            for elem in elems:
                self.validate(elem.value)
                elem.delete()
        return
       

    @db_session
    def __delattr__(self, key):
        delete(e for e in self.store if e.key == key)


    @db_session
    def add(self, value, key=None):
        self.validate(value)
        hex = uuid.uuid1().hex
        key = "_{}".format(hex) if not isinstance(key, str) else key
        self.store(key=key, value=value)
        return key

    @db_session
    def query_key(self, key, for_update=False):
        elem = None
        if for_update:
            elem = select(e for e in self.store if e.id == self.id).for_update().first()
        else:
            elem = select(e for e in self.store if e.id == self.id).first()
        if elem:
            self.validate(elem.value)
            return StoreMeta(elem, store=self.store)

    @db_session
    def query_meta(self, key):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key, "meta")
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=False)
        for elem in elems:
            self.validate(elem.value, extra=elem.key)
        return StoreMetas(elems, store=self.store)

    @db_session
    def queryx(self, key, x):
        if isinstance(key, slice):
            raise StoreException('not implemented!')
        elif isinstance(key, tuple):
            key='.'.join(key)

        # string key
        filters = parse(key, x)
        elems = select(e for e in self.store)
        if filters:
            elems = elems.filter(filters)
        if self.order == 'desc':
            elems = elems.order_by(lambda o: desc(o.create)).order_by(lambda o: desc(o.id))
        elems = self.adjust_slice(elems, for_update=False)
        for elem in elems:
            self.validate(elem.value, extra=elem.key)
        return StoreMetas(elems, store=self.store)
    def adjust_slice(self, elems, for_update=False):
        if for_update:
            elems = elems.for_update()
        begin, end = self.begin, self.end
        if begin and end:
            # pony doesn't support step here
            if begin < 0:
                begin = len(self) + begin
            if end < 0:
                end = len(self) + end
            if begin > end:
                begin, end = end, begin
            elems = elems[begin:end]
        elif begin:
            if begin < 0:
                begin = len(self) + begin
            elems = elems[begin:]
        elif end:
            if end < 0:
                end = len(self) + end
            elems = elems[:end]
        else:
            elems = elems[:]
        return elems

    @db_session
    def __len__(self):
        return count(e for e in self.store)
